/*  This file declares the possible device IO control commands for the
    apanel device
 
    Copyright (C) 2001 Willem van Engen <wvengen@stack.nl> 
    Copyright (C) 2001 Jochen Eisinger <jochen.eisinger@gmx.net>

 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * ALTERNATIVELY, this product may be distributed under the terms of
 * the GNU Public License, in which case the provisions of the GPL are
 * required INSTEAD OF the above restrictions.  (This clause is
 * necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)     
 * OF THE POSSIBILITY OF SUCH DAMAGE.   
*/

#ifndef _APANEL_H
#define _APANEL_H

#ifdef __FreeBSD__		/* Mind the case! */
#  define APANEL_MAJOR 127	/* Until there is an official freebsd major */
#  define APANEL_DEV	"/dev/apanel"
#elif defined __linux__
#  include <linux/ioctl.h>
#  define APANEL_MAJOR 10
#  define APANEL_MINOR 216
#  define APANEL_DEV	"/dev/fujitsu/apanel"
#endif

/* Describes to which subsystem the ioctl applies. 'P' for proprietary? */
#define APANEL_IOCTL_GROUP	'P'

/*
 * Driver interface version
 * Returns 100 for this interface
 */
#define IOCTL_APANEL_VER	_IOR(APANEL_IOCTL_GROUP,200,int)

#define APANEL_VER_MAJOR(ver)	((ver) >> 8)
#define APANEL_VER_MINOR(ver)	((ver) & 0xFF)

/*
 * General capabilities (which devices exist)
 *
 * if (ioctl(IOCTL_APANEL_CAP_GEN)&APANEL_CAP_GEN_FOO ) 
 *                      then capability FOO is supported
 *
 */
#define APANEL_CAP_GEN_LED	1
#define APANEL_CAP_GEN_LCD	2
#define APANEL_CAP_GEN_APPBTN	4
#define APANEL_CAP_GEN_CDBTN	8
#define IOCTL_APANEL_CAP_GEN	_IOR(APANEL_IOCTL_GROUP,201,int)

/*
 * Use one ioctl for all devices
 * This has the advantage that one interface is defined
 * for all devices. The caps can tell what is possible on
 * a certain device (with some defaults for each device
 * of course which the programmer doesn't have to check
 * explicitly like on/off for led and read for buttons).
 */

struct apanelcmd {
  int device;		/* Device to operate command on (APANEL_DEV_GEN_*) */
  int cmd;		/* Command to execute (APANEL_CMD_*) */
  int data;		/* Data for command */
};

#define IOCTL_APANEL_CMD	_IOR(APANEL_IOCTL_GROUP,202,struct apanelcmd)

/* Devices */
#define APANEL_DEV_LED		1
#define APANEL_DEV_LCD		2
#define APANEL_DEV_CDBTN	3
#define APANEL_DEV_APPBTN	4

/*
 * Commands
 */

/* Device capabilities */
#define APANEL_CMD_CAPS		1

#define APANEL_CAP_ONOFF	1	/* Turn on/off */
#define APANEL_CAP_GET		2	/* General get command */
#define APANEL_CAP_SET		4	/* General set command */
#define APANEL_CAP_BLINK	8	/* Blinking */
#define APANEL_CAP_DUTY		0x10	/* Duty cycle of blinking */
#define APANEL_CAP_RESET	0x20	/* Reset device */
#define APANEL_CAP_MAX		0x40	/* Get highest possible value for */
					/* set/get */
#define APANEL_CAP_MIN		0x80	/* Get lowest possible value */

#define APANEL_CMD_ONOFF	2
#define APANEL_ONOFF_ON		1	/* Or is APANEL_ON nicer? */
#define APANEL_ONOFF_OFF	0

#define APANEL_CMD_SET		3
#define APANEL_CMD_GET		4

#define APANEL_CMD_BLINK	5	/* data=Blink frequency *0.01Hz or so */
#define APANEL_CMD_DUTY		6	/* data=percentage high */

#define APANEL_CMD_RESET	7	/* If this is useful at all */

#define APANEL_CMD_MAX		8
#define APANEL_CMD_MIN		9

/*
 * Button masks
 */
/* Masks for application buttons */
#define APANEL_APPBTN_A		1
#define APANEL_APPBTN_B		2
#define APANEL_APPBTN_INTERNET	4
#define APANEL_APPBTN_EMAIL	8

/* Masks for cd buttons */
#define APANEL_CDBTN_STOP	1
#define APANEL_CDBTN_EJECT	2
#define APANEL_CDBTN_PLAY	4
#define APANEL_CDBTN_PAUSE	8
#define APANEL_CDBTN_BACK	0x10
#define APANEL_CDBTN_FORW	0x20

#endif /* __APANEL_H__ */

