/*
 * Fujitsu c4110 application panel driver
 * apanelc.c - the userland control program
 *
 * (c) Copyright by Willem van Engen
 * See COPYRIGHT for copyright
 * See CREDITS for credits
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/syscall.h>
#include <string.h>
#include <err.h>
#include "apanel.h"

#define atoix(A) (int)strtol(A,(char**)NULL,0)

int fd;

void apanel_set(int, int, int);
int apanel_get(int, int);

int main(int argc,char* argv[]) {
  int i,stat;

  if ((fd=open(APANEL_DEV,O_RDWR)) < 0) {
    err(1,APANEL_DEV);
  }

  if (argc==1) goto usage;
  for (i=1;i<argc;i++) {
    /********* LED *********/
    if (strcasecmp(argv[i],"led")==0) {
      if (argc-i<=1) goto ledusage;
      if (strcasecmp(argv[++i],"on")==0) {
        stat=1;
	apanel_set(APANEL_DEV_LED,APANEL_CMD_ONOFF,APANEL_ONOFF_ON);
      } else if(strcasecmp(argv[i],"off")==0) {
        stat=0;
	apanel_set(APANEL_DEV_LED,APANEL_CMD_ONOFF,APANEL_ONOFF_OFF);
      } else if(strcasecmp(argv[i],"blink")==0) {
        if (argc-i<=1) goto blinkusage;
        stat=atoix(argv[++i]);
	if (stat>=apanel_get(APANEL_DEV_LED,APANEL_CMD_MIN) && 
		stat<=apanel_get(APANEL_DEV_LED,APANEL_CMD_MAX) ) {
	  apanel_set(APANEL_DEV_LED,APANEL_CMD_BLINK,stat);
        } else {
blinkusage:
          fprintf(stderr,"Blink requires argument between 0x10 and 0x7f.\n");
	  return 3;
        }
      } else if(strcasecmp(argv[i],"stat")==0 || 
               strcasecmp(argv[i],"status")==0) {
        printf("%d\n",apanel_get(APANEL_DEV_LED,APANEL_CMD_GET));
      } else {
ledusage:
        fprintf(stderr,"usage: apanelc led on|off|blink xx|stat[us]\n");
	return 3;
      }
    /********* LCD *********/
    } else if(strcasecmp(argv[i],"lcd")==0) {
      if (argc-i<=1) goto lcdusage;
      if (strcasecmp(argv[++i],"stat")==0 ||
                 strcasecmp(argv[i],"status")==0) {
	printf("%d\n",apanel_get(APANEL_DEV_LCD,APANEL_CMD_GET));
      } else if (strcasecmp(argv[i],"off")==0) {
	  apanel_set(APANEL_DEV_LCD,APANEL_CMD_ONOFF,APANEL_ONOFF_OFF);
      } else if (argv[i][0]=='-' && (argv[i][1]=='\0' ||
                  (argv[i][1]=='-'&&argv[i][2]=='\0')) ) {
	  apanel_set(APANEL_DEV_LCD,APANEL_CMD_SET,
		  apanel_get(APANEL_DEV_LCD,APANEL_CMD_MAX)+1 );
      } else {
        stat=atoix(argv[i]);
        if ( stat>apanel_get(APANEL_DEV_LCD,APANEL_CMD_MAX) ||
	  stat<apanel_get(APANEL_DEV_LCD,APANEL_CMD_MIN) ) {
lcdusage:
            fprintf(stderr,"usage: apanelc lcd stat[us]|off|-[-]|xxx\n"
                         "  xxx  number between 0 and 99 to put on lcd\n");
	    return 3;
        }
	apanel_set(APANEL_DEV_LCD,APANEL_CMD_ONOFF,APANEL_ONOFF_ON);
	apanel_set(APANEL_DEV_LCD,APANEL_CMD_SET,stat);
      }
    /******** Wrong device *********/
    } else {
usage:
      fprintf(stderr,"usage: apanelc led stat[us]|on|off|blink xxx\n"
                     "       apanelc lcd stat[us]|off|-[-]|xxx\n");
      return 3;
    }
  }

  close(fd);
  return 0;
}

void apanel_set(int device, int cmd, int data) {
  struct apanelcmd apcmd={device,cmd,data};
  if (ioctl(fd,IOCTL_APANEL_CMD,&apcmd)==-1)
    err(1,"IOCTL_APANEL_CMD(0x%x,0x%x,0x%x)",device,cmd,data);
}
int apanel_get(int device, int cmd) {
  struct apanelcmd apcmd={device,cmd,0};
  int result;
  if ((result = ioctl(fd,IOCTL_APANEL_CMD,&apcmd))==-1)
    err(1,"IOCTL_APANEL_CMD(0x%x,0x%x, )",device,cmd);
  return result;
}

