/*
    FJKeyInf client
    
    Copyright (C) 2001 Jochen Eisinger <jochen.eisinger@gmx.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>
#include <pwd.h>
#include <sys/kd.h>

/* #include <linux/apanel.h> */
#include "apanel.h"

#define VERSION	"fjkeyd - Fujitsu Application Panel Key Daemon v0.1"

#define FJKEYD_PID_FILE	"/var/run/fjkeyd.pid"

#define FJKEYD_MODE_DAEMON	0
#define FJKEYD_MODE_RELOAD	1
#define FJKEYD_MODE_STATUS	2
#define FJKEYD_MODE_STOP	3
#define FJKEYD_MODE_HELP	4
#define FJKEYD_MODE_VERSION	5
#define FJKEYD_MODE_INVAL	-1


static int	prog_mode = FJKEYD_MODE_DAEMON;

static int	daemon_pid = -1;

void parse_options(int argc, char ** args)
{
    if (argc == 1)
	return;

    if (argc > 2) {
	prog_mode = FJKEYD_MODE_INVAL;
	return;
    }

    if (strcmp(args[1],"-r") == 0) {
	prog_mode = FJKEYD_MODE_RELOAD;
	return;
    }

    if (strcmp(args[1],"-s") == 0) {
	prog_mode = FJKEYD_MODE_STATUS;
	return;
    }

    if (strcmp(args[1],"-k") == 0) {
	prog_mode = FJKEYD_MODE_STOP;
	return;
    }

    if (strcmp(args[1],"-h") == 0) {
	prog_mode = FJKEYD_MODE_HELP;
	return;
    }

    if (strcmp(args[1],"-v") == 0) {
	prog_mode = FJKEYD_MODE_VERSION;
	return;
    }

    prog_mode = FJKEYD_MODE_INVAL;
}

void get_daemon_pid(void)
{
    FILE	*fp;
    struct	stat buf;
    char	tmp[128];

    if ((fp = fopen(FJKEYD_PID_FILE, "r+")) == NULL)
	return;

    fscanf(fp, "%d", &daemon_pid);

    fclose(fp);

    sprintf(tmp, "/proc/%d", daemon_pid);

    if (stat(tmp, &buf) < 0) {
	remove(FJKEYD_PID_FILE);
	daemon_pid = -1;
    }
}

void set_daemon_pid(void)
{
    FILE	*fp;

    if ((fp = fopen(FJKEYD_PID_FILE, "w+")) == NULL) {
	fprintf(stderr, "could not write pid-file...\n");
	return;
    }

    fprintf(fp, "%d", daemon_pid);

    fclose(fp);
}

void test_fjkeyinf_dev(void)
{
    int	fdev;

    if ((fdev = open(APANEL_DEV, O_RDONLY)) < 0) {
	fprintf(stderr, "could not find apanel device...\n");
	exit(1);
    }
    close(fdev);
}

void do_key(char **args) {
    int fd, mode, console;
    char tty[16], buf[16];
    struct stat fs;
    struct passwd *pw;

    if (fork() == 0) {
	close (0);
	close (1);
	close (2);
	if ((fd = open("/dev/tty0", O_RDONLY)) < 0)
	    goto do_exec;

	mode = 12; /* Get foreground console */
	if ((console = ioctl(fd, TIOCLINUX, &mode)) < 0) {
	    close(fd);
	    goto do_exec;
	}

	close(fd);
	sprintf (tty, "/dev/tty%u", console + 1);
	
	if ((fd = open(tty, O_RDONLY)) < 0)
	    goto do_exec;

	mode = 6; /* Get shift state */
	if (ioctl(fd, TIOCLINUX, &mode) >= 0) {
	    sprintf (buf, "%u", mode);
	    setenv ("KD_SHIFT", buf, 1);
	}
	fstat(fd, &fs);
	if (ioctl(fd, KDGETMODE, &mode) >= 0)
	    setenv ("KD_MODE", mode == KD_GRAPHICS ? "GRAPHICS" : "TEXT",1);

	if (mode == KD_GRAPHICS) {
	    close (fd);
	    if (fs.st_uid == 0)		/* XDM */
		stat("/dev/console", &fs);
	} else {
	    open(tty, O_WRONLY);
	    dup(1);
	}
	sprintf (buf, "%u", fs.st_uid);
	setenv ("KD_UID", buf, 1);
	if ((pw = getpwuid (fs.st_uid)) != NULL)
		setenv ("KD_USER", pw->pw_name, 1);

do_exec:
	execvp(args[0], args);

	exit(0);
    }
}

void poll_fjkeyinf(void)
{
    FILE	*config;
    int	fdev, ctr, i, l;
    int	state;
    char	tmp[256], *p;
    int	argc[4] = {0,0,0,0};
    char	**args[4] = {NULL, NULL, NULL, NULL};
    struct	apanelcmd cmd;

    if ((config = fopen("/etc/fjkeyd.conf", "r")) == NULL) {
		fprintf(stderr, "could not read /etc/fjkeyd.conf...\n");
		exit(1);
    }

	
    for (ctr=0 ; ctr<4 ; ctr++) {

	fgets(tmp, 255, config);
	
	tmp[255] = 0;

	if ((strlen(tmp)) && (tmp[strlen(tmp)-1] == '\n')) {
	    tmp[strlen(tmp)-1] = 0;
	}
	
	p = tmp;
	argc[ctr] = 1;

	while (*p != 0) {
	    if (*p == '\t')
		argc[ctr]++;
	    p++;
	}

	args[ctr] = malloc(sizeof(char *) * (argc[ctr] + 1));
	
	p = tmp;

	memset(args[ctr], 0, sizeof(char *) * (argc[ctr] + 1));

	for (i = 0; i<argc[ctr] ; i++) {

	    if (i == (argc[ctr] - 1))
		l = strlen(p);
	    else
		l = (strchr(p, '\t') - p);

	    (args[ctr])[i] = malloc(l + 1);

	    memcpy((args[ctr])[i], p, l);

	    ((args[ctr])[i])[l] = 0;
		
	    p += l;
	    p++;
	}
    }

    unsetenv ("KD_SHIFT");
    unsetenv ("KD_MODE");
    unsetenv ("KD_UID");
    unsetenv ("KD_USER");

    fclose(config);

    fdev = open(APANEL_DEV, O_RDONLY);

    cmd.device = APANEL_DEV_APPBTN;
    cmd.cmd = APANEL_CMD_GET;

    for (;;) {
	state = ioctl(fdev, IOCTL_APANEL_CMD, (unsigned long)&cmd);

	if (state & APANEL_APPBTN_A)
	    do_key(args[0]);

	if (state & APANEL_APPBTN_B)
	    do_key(args[1]);

	if (state & APANEL_APPBTN_INTERNET)
	    do_key(args[2]);

	if (state & APANEL_APPBTN_EMAIL)
	    do_key(args[3]);

	usleep(500);

	waitpid(-1, NULL, WNOHANG);
    }
    close(fdev);
}

int main(int argc, char ** args)
{
    parse_options(argc, args);

    switch (prog_mode) {
	case FJKEYD_MODE_INVAL:
	    fprintf(stderr,"%s: invalid arguments. Use -h for help.\n",args[0]);
	    return 1;

	case FJKEYD_MODE_HELP:
	    fprintf(stdout, VERSION "\n");
	    fprintf(stdout, "\n%s: [option]\n", args[0]);
	    fprintf(stdout, "\n\t-h\tshows this help"
			"\n\t-k\tkills running daemon"
			"\n\t-r\treloads configuration file"
			"\n\t-s\tdisplays status information"
			"\n\t-v\tdisplay version"
			"\n\nWhen invoked without any options "
			"a new daemon is started if possible\n" );
	    return 0;

	case FJKEYD_MODE_VERSION:
	    fprintf(stdout, VERSION "\n");
	    return 0;
    }

    if (getuid() != 0) {
	fprintf(stderr, "%s: you have to be root\n", args[0]);
	return 1;
    }

    test_fjkeyinf_dev();

    get_daemon_pid();

    switch (prog_mode) {
	case FJKEYD_MODE_STATUS:
	    if (daemon_pid == -1) {
		fprintf(stdout, "%s: daemon not present\n", args[0]);
	    } else {
		fprintf(stdout, "%s: daemon active (pid: %d)\n",
				args[0], daemon_pid);
	    }
	    return 0;

	case FJKEYD_MODE_STOP:
	    if (daemon_pid == -1) {
		fprintf(stderr, "%s: daemon not present\n", args[0]);
		return 1;
	    }

	    kill(daemon_pid, SIGTERM);
	    remove(FJKEYD_PID_FILE);

	    return 0;
			
	case FJKEYD_MODE_RELOAD:
	    if (daemon_pid != -1) {
		kill(daemon_pid, SIGTERM);
		remove(FJKEYD_PID_FILE);
	    }
    }

    switch (daemon_pid = fork()) {
	case -1:
	    fprintf(stderr, "%s: fork() failed...\n", args[0]);
	    return 1;

	case 0:
	    poll_fjkeyinf();
	    return 0;
    }

    set_daemon_pid();

    return 0;
}
