/*
    SMBus client for the Fujitsu Siemens Lifebook C-6535 Application Panel
    
    Copyright (C) 2001 Jochen Eisinger <jochen.eisinger@gmx.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/ctype.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/ioport.h>
#include <linux/types.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/i2c.h>
#include <linux/i2c-id.h>
#include <linux/miscdevice.h>

/*
 * in this file are the constants definied, an apanel driver should
 * understand
 */
/* #include <linux/apanel.h> */
#include "apanel.h"

/* print lots of useless debug infos */
#define DEBUG 0

/*
 * this is the internal driver version number. There is no direct relationship
 * between the hardware, the apanel protocol, or the driver versions
 * However, this version number should be increased for every change
 */
/* current driver version */
#define FJKEYINF_VERSION_MAJOR	0
#define FJKEYINF_VERSION_MINOR	4

/*
 * This is the apanel version this driver implements. This _must_ be the
 * same as your apanel.h file
 */
/* protocol version */
#define FJKEYINF_APANEL_MAJOR	1
#define FJKEYINF_APANEL_MINOR	0

/*
 * Some real leet m4cr0s
 */
#define STRINGIFY1(x)	#x
#define STRINGIFY(x)	STRINGIFY1(x)

#define PASTE1(x,y)	x##y
#define PASTE(x,y)	PASTE1(x,y)

/*
 * This is the version of the driver as a string
 */
#define FJKEYINF_VERSION_STRING	PASTE("v",				\
				PASTE(STRINGIFY(FJKEYINF_VERSION_MAJOR),\
				PASTE(".",STRINGIFY(FJKEYINF_VERSION_MINOR))))

/*
 * This is the version of the protocol as a string
 */
#define FJKEYINF_APANEL_VERSION_STRING	PASTE("v",			\
					PASTE(STRINGIFY(FJKEYINF_APANEL_MAJOR),\
					PASTE(".",			\
					STRINGIFY(FJKEYINF_APANEL_MINOR))))

/* 
 * every i2c device has to register to the i2c sub-system. Therefor it needs
 * a driver id. We should be I2C_DRIVERID_FJKEYINF. However, if this isn't
 * defined in i2c-id.h, we'll use a generic id.
 */
#ifndef I2C_DRIVERID_FJKEYINF
#define I2C_DRIVERID_FJKEYINF	0xF000
#endif

/*
 * Yes, it's unbelievable, but me crappy driver has an official devices
 * entry. It's register as a misc-device (char-major-10) minor 216. The
 * location should be /dev/fujitsu/apanel... after all, it seems to be
 * a quite c00l driver ;>
 */
#define FJKEYINF_CHAR_MINOR	216

/*
 * Modules can store nice extra infos...
 */
#ifdef MODULE
MODULE_AUTHOR("Jochen Eisinger <jochen.eisinger@gmx.net>");
MODULE_DESCRIPTION("Application Panel driver for Lifebook C-series");
MODULE_LICENSE("GPL");
#endif

/*
 * So, let's start the hackin'
 *
 * we first define the interface for the i2c driver, cuz the misc device
 * part uses the struct fjkeyinf_client. also note that this struct is
 * static where it would be better to dynamically allocate one... but what
 * fsck, more than one apanel your crappy laptop won't have...
 */

/* definitions for i2c (smbus) interface */

/* forward declaration of the interface procedures */

/* detach removes the smbus driver from the system */
	static int fjkeyinf_detach(struct i2c_client *client);
/* attach tries to attach to a given smbus */
	static int fjkeyinf_attach(struct i2c_adapter *adap);
/* these two functions are called if someone "uses" us or stops to do so */
	void fjkeyinf_dec_use(struct i2c_client *client);
	void fjkeyinf_inc_use(struct i2c_client *client);

/* this structur defines the interface to the i2c sub-system */
static struct i2c_driver fjkeyinf_driver = {

	/* name		*/ "Application Panel driver " FJKEYINF_VERSION_STRING,
	/* id		*/ I2C_DRIVERID_FJKEYINF,
	/* flags	*/ I2C_DF_NOTIFY,
	/* attach	*/ &fjkeyinf_attach,
	/* detach	*/ &fjkeyinf_detach,
	/* command	*/ NULL,
	/* inc_use	*/ &fjkeyinf_inc_use,
	/* dec_use	*/ &fjkeyinf_dec_use
};


/* Addresses to scan. afaik the device is at id 0x18. so we only scan this */
static unsigned short normal_i2c[] = {0,I2C_CLIENT_END};
static unsigned short normal_i2c_range[] = {I2C_CLIENT_END};

/* 
 * generate some stupid additional structures. we don't really care about
 * this
 */
I2C_CLIENT_INSMOD;

/*
 * If we've found the device, we have to provide an i2c_client structure
 * to the i2c subsystem with which this devices can be accessed. as I stated
 * above, there's max. 1 device, so I use _one static_ entry
 */
static struct i2c_client fjkeyinf_client = 
{
        "fjkeyinf",
        -1,
        0,
        0,
        NULL,
        &fjkeyinf_driver,
	NULL
};

static int led_type, btn_type;

/*
 * luckily, the c-series laptops have a configuration block in there BIOS.
 * so we can use it to detect the presence of the apanel device. There's
 * also a checksum somewhere, but we don't care about it.
 *
 * Note the first 8 characters. that's where this strange driver name comes
 * from.
 *
 * The configuration block can be found at 0x000FFA30
 *
 * There should also be an access method 3, but I don't know much about it.
 * Basically there should also be an LCD device. but my notebook hasn't one
 * so I don't know what the configuration block looks like
 *
 * type 4 is LED type 1 is BUTTONS, 7 is 3 BUTTONS, probably 2 is LCD.
 */
static const unsigned char fjkeyinf_info[16] = {
	'F', 'J', 'K', 'E', 'Y', 'I', 'N', 'F' };
	/* device 0        type		  	1,
			   access method  	1,
			   enabled	  	1,
			   smbus device	  	0x30,
	   device 1        type		  	4,
			   access method  	1,
			   enabled	  	1,
			   smbus device   	0x30}; */

/* Highest one should be first */
static unsigned fjkeyinf_addr[] = { 0x000FFA30, 0x000F6F70, 0 };

/*
 * the following functions implement the ioctls. Note however, that not
 * much is implemented yet.
 */

/* Get the capabilities for a specific device */
int fjkeyinf_caps(int dev)
{
    switch (dev) {
	case APANEL_DEV_LED:
	    if (!led_type)
		return -EINVAL;
	    return (APANEL_CAP_ONOFF + APANEL_CAP_RESET + APANEL_CAP_GET);

	case APANEL_DEV_APPBTN:
	    if (!btn_type)
		return -EINVAL;
	    return APANEL_CAP_GET;

	default:
#if DEBUG > 0
	    printk(KERN_NOTICE "fjkeyinf: request device"
			" capabilities for unknown device %d\n", dev);
#endif
	    return -EINVAL;
    }
}

/* turn a device on or off */

int fjkeyinf_onoff(int dev, int state)
{
    switch (dev) {
	case APANEL_DEV_LED:
	    if (!led_type)
		return -EINVAL;
#if DEBUG > 1
	    printk(KERN_DEBUG "fjkeyinf: turning LED %s...\n",state?"on":"off");
#endif
	    i2c_smbus_write_word_data(&fjkeyinf_client, 0x10, state?0x8080:0);
	    return state;

	default:
#if DEBUG > 0
	    printk(KERN_NOTICE "fjkeyinf: ONOFF called for invalid device"
			" (dev %d, state %d)\n", dev, state);
#endif
	    return -EINVAL;
    }
}

/* sets a device to a specific value */
int fjkeyinf_set(int dev, int data)
{
    switch (dev) {
	default:
#if DEBUG > 0
	    printk(KERN_NOTICE "fjkeyinf: SET called for invalid device"
		    " (dev %d, data %d)\n", dev, data);
#endif
	    return -EINVAL;
    }

}

/* gets the current value from a device */
int fjkeyinf_get(int dev)
{

    switch (dev) {
	case APANEL_DEV_LED:
	    switch (led_type) {
		default:
		    return -EINVAL;
		case 1:
		    return(i2c_smbus_read_word_data(&fjkeyinf_client,0)>>7)&1;
		case 4:
		    return(i2c_smbus_read_word_data(&fjkeyinf_client,16)>>15)&1;
	    }

	case APANEL_DEV_APPBTN:
	    if (btn_type) { /* ARREGLAME */
		int state;

		if (btn_type == 1)
			state = i2c_smbus_read_word_data(&fjkeyinf_client, 0);
		else
			state = i2c_smbus_read_word_data(&fjkeyinf_client, 8);

		if (state == 0xffffffff)
		    return 0;
	
		if (btn_type == 1)
		    state = ((state >> 8) & 0x0f) ^ 0x0f;
		else
		    i2c_smbus_write_word_data(&fjkeyinf_client, 0x08, 0);

		return (((state & 0x1) ? APANEL_APPBTN_EMAIL : 0)
			| ((state & 0x2) ? APANEL_APPBTN_INTERNET : 0)
			| ((state & 0x4) ? APANEL_APPBTN_A : 0)
			| ((state & 0x8) ? APANEL_APPBTN_B : 0));

	    }

	default:
#if DEBUG > 0
	    printk(KERN_NOTICE "fjkeyinf: GET called for invalid"
				" device (dev %d)\n", dev);
#endif
	    return -EINVAL;

    }

}

/* sets the blink frequency of a device */
int fjkeyinf_blink(int dev, int freq)
{
    switch (dev) {
	case APANEL_DEV_LED:
	    if (!led_type)
		return -EINVAL;
	    i2c_smbus_write_word_data(&fjkeyinf_client, 0x10, 0x8000);
	    return 0;

	default:
#if DEBUG > 0
	    printk(KERN_NOTICE "fjkeyinf: BLINK called for invalid"
				" device (dev %d, freq %d)\n", dev, freq);
#endif
	    return -EINVAL;
    }
}

/* 
 * Honestly, I don't what this ioctl should do, maybe I'll look it up in a
 * dictionary some day. But at least til then this function will just
 * return ENOCLUE
 */
int fjkeyinf_duty(int dev, int data)
{
    switch (dev) {
	default:
#if DEBUG > 0
	    printk(KERN_NOTICE "fjkeyinf: DUTY called for invalid device "
		    "(dev %d, data %d) returning -ENOCLUE\n", dev, data);
#endif
	    return -EINVAL;

    }
}

/* resets a device */
int fjkeyinf_reset(int dev)
{
    switch (dev) {
	case APANEL_DEV_LED:
	    if (!led_type)
		return -EINVAL;
	    fjkeyinf_onoff(dev, 0);
	    return 0;

	default:
#if DEBUG > 0
	    printk(KERN_NOTICE "fjkeyinf: can't reset this device"
				" (dev %d)\n", dev);
#endif
	    return -EINVAL;

    }

}

/* returns the maximal value for SET */
int fjkeyinf_max(int dev)
{
    switch (dev) {
	case APANEL_DEV_LED:
	    if (!led_type)
		return -EINVAL;
	    return 100;

	default:
#if DEBUG > 0
	    printk(KERN_NOTICE "fjkeyinf: MAX called for invalid"
				" device (dev %d)\n", dev);
#endif
	    return -EINVAL;
    }
}

/* returns the minimal value for SET */
int fjkeyinf_min(int dev)
{
    switch (dev) {
	case APANEL_DEV_LED:
	    if (!led_type)
		return -EINVAL;
	    return 100;

	default:
#if DEBUG > 0
	    printk(KERN_NOTICE "fjkeyinf: MIN called for invalid"
				" device (dev %d)\n", dev);
#endif
	    return -EINVAL;
    }
}


/* 
 * file operations for /dev/fujitsu/apanel
 * 
 * That what the name says: file operation. pretty basic stuff follows
 */
			
/*
 * if somebody opens us, we just realize and let him pass
 */
static int fjkeyinf_open (struct inode *inode, struct file *filp)
{
#if DEBUG > 1
    printk(KERN_DEBUG "fjkeyinf: open called for /dev/fujitsu/apanel\n");
#endif
    MOD_INC_USE_COUNT;
    return 0;
}

/* same with close */
static int fjkeyinf_close (struct inode *inode, struct file *filp)
{
#if DEBUG > 1
    printk(KERN_DEBUG "fjkeyinf: close called for /dev/fujitsu/apanel\n");
#endif
    MOD_DEC_USE_COUNT;
    return 0;
}

/* all commands are passed using the ioctl interface. so we have to decide
 * here, what we have to do */
int fjkeyinf_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
		unsigned long args)
{
    struct apanelcmd *acmd;

#if DEBUG > 1
    printk(KERN_DEBUG "fjkeyinf: ioctl (cmd: %u, args: 0x%08lx)\n",
			cmd, args);
#endif

	/* Let's see, what they want from us... */
	
    switch (cmd) {
	case IOCTL_APANEL_VER:

	    /* this ioctl returns the protocol version the driver supports */
#if DEBUG > 1
	    printk(KERN_DEBUG "fjkeyinf: apanel version " 
				FJKEYINF_APANEL_VERSION_STRING "\n");
#endif
	    return ((FJKEYINF_APANEL_MAJOR << 8) + FJKEYINF_APANEL_MINOR);

	case IOCTL_APANEL_CAP_GEN:
	    /* this ioctl returns all supported devices. Although the apanel
	     * protocol defines lots of devices, not every driver has to
	     * support all of them... more devices will follow */
#if DEBUG > 1
	    printk(KERN_DEBUG "fjkeyinf: capabilities = LED, APPBTN\n");
#endif
	    return (APANEL_CAP_GEN_LED | APANEL_CAP_GEN_APPBTN);

	case IOCTL_APANEL_CMD:
	    /* the actual command is passed in a apanelcmd structure */

	    if (!args) {
#if DEBUG > 0
		printk(KERN_NOTICE "fjkeyinf: invalid apanel command "
						"(NULL pointer)\n");
#endif
		return -EINVAL;
	    }

	    acmd = (struct apanelcmd *)args;

	    /* although not all commands are implemented, we understand all.. */

#if DEBUG > 1
	    printk(KERN_DEBUG "fjkeyinf: apanel command %d\n", acmd->cmd);
#endif
	    switch (acmd->cmd) {
		case APANEL_CMD_CAPS:
		    return fjkeyinf_caps(acmd->device);

		case APANEL_CMD_ONOFF:
		    return fjkeyinf_onoff(acmd->device, acmd->data);

		case APANEL_CMD_SET:
		    return fjkeyinf_set(acmd->device, acmd->data);

		case APANEL_CMD_GET:
		    return fjkeyinf_get(acmd->device);

		case APANEL_CMD_BLINK:
		    return fjkeyinf_blink(acmd->device, acmd->data);

		case APANEL_CMD_DUTY:
		    return fjkeyinf_duty(acmd->device, acmd->data);

		case APANEL_CMD_RESET:
		    return fjkeyinf_reset(acmd->device);

		case APANEL_CMD_MAX:
		    return fjkeyinf_max(acmd->device);

		case APANEL_CMD_MIN:
		    return fjkeyinf_min(acmd->device);

		default:
#if DEBUG > 0
		    printk(KERN_NOTICE "fjkeyinf: unknown device/command"
				       " %d/%d\n", acmd->device, acmd->cmd);
#endif
		    return -EINVAL;
		}

	default:
#if DEBUG > 0
	    printk(KERN_NOTICE "fjkeyinf: unknown ioctl code %u\n", cmd);
#endif
	    return -EINVAL;
    }
}

/* now we tell the misc_device what nice functions we've implemented */
static struct file_operations fjkeyinf_fops = {
	ioctl:		fjkeyinf_ioctl,
	open:		fjkeyinf_open,
	release:	fjkeyinf_close
};

/* misc dev entry. We need this to register to the misc dev driver */
static struct miscdevice fjkeyinf_dev = {
	FJKEYINF_CHAR_MINOR,
	"apanel",
	&fjkeyinf_fops
};


/* Now comes the i2c driver stuff. This is pretty straight forward, just as they
 * described it in their docu.
 */

/* basically this function should probe the i2c client, but we know that it has
 * to be the one we're looking for - and I have no idea how I should probe for
 * it, so we just register... */
static int fjkeyinf_probe(struct i2c_adapter *adap, int addr,
		     unsigned short flags, int kind)
{
    int result;
    fjkeyinf_client.adapter = adap;
    fjkeyinf_client.addr = addr;

    if ((result = misc_register(&fjkeyinf_dev)) < 0) {
#if DEBUG > 0
	printk(KERN_NOTICE "fjkeyinf: could not register misc device (%d)\n",
				result);
#endif
	return result;

    }

    i2c_attach_client(&fjkeyinf_client);

    printk(KERN_INFO "fjkeyinf: Application Panel Driver "
		       	FJKEYINF_VERSION_STRING "\n");
    printk(KERN_INFO "fjkeyinf: Copyright (c) 2001 by "
		"Jochen Eisinger <jochen.eisinger@gmx.net>\n");
	
#if DEBUG > 1
    printk(KERN_DEBUG "fjkeyinf: driver loaded at address 0x%02x...\n", addr);
#endif
    return 0;
}


/* this function is invoked, when we should release our resource... */
static int fjkeyinf_detach(struct i2c_client *client)
{
#if DEBUG > 1
    printk(KERN_DEBUG "fjkeyinf: driver detached...\n");
#endif
    i2c_detach_client(client);
	
    misc_deregister(&fjkeyinf_dev);

    return 0;
}

/* this function is invoked for every i2c adapter, that has a device at the
 * address we've specified */
static int fjkeyinf_attach(struct i2c_adapter *adap)
{
	/* if this adapter is a PIIX4, we just attach to it */
    if ((adap->id == (I2C_ALGO_SMBUS | I2C_HW_SMBUS_PIIX4)) ||
       (adap->id == (I2C_ALGO_SMBUS | I2C_HW_SMBUS_ALI1535)))
	    return i2c_probe(adap, &addr_data, fjkeyinf_probe);

return 0;
}

/* the next to functions control how many ppl are accessing us */
void fjkeyinf_inc_use(struct i2c_client *client)
{
#ifdef MODULE
  MOD_INC_USE_COUNT;
#endif
}

void fjkeyinf_dec_use(struct i2c_client *client)
{
#ifdef MODULE
  MOD_DEC_USE_COUNT;
#endif
}

/* startup */
#ifdef MODULE
int init_module(void)
#else
int fjkeyinf_init(void)
#endif
{
    unsigned char *fujitsu_bios;
    int i = 0;

    if (__pa(high_memory) < (fjkeyinf_addr[0] - 16)) {
#if DEBUG > 1
	printk(KERN_DEBUG "fjkeyinf: Fujitsu BIOS not found...\n");
#endif
	return 0;
    }            

    do {
	fujitsu_bios = __va(fjkeyinf_addr[i]);
    } while ((memcmp(fujitsu_bios, fjkeyinf_info, 8) != 0) &&
						fjkeyinf_addr[++i]);

    if (memcmp(fujitsu_bios, fjkeyinf_info, 8) != 0) {
#if DEBUG > 1
	printk(KERN_DEBUG "fjkeyinf: device not found...\n");
#endif
	return 0;
    }

    normal_i2c[0] = fujitsu_bios[11] / 2;

    for (i = 8; fujitsu_bios[i]; i += 4) {
	if (normal_i2c[0] != fujitsu_bios[i + 3] / 2) {
	    printk(KERN_WARNING "fjkeyinf: Only one SMBus address supported. "
				"Using %x, ignoring %x\n", normal_i2c[0],
				fujitsu_bios[i + 3] / 2);
	} else switch (fujitsu_bios[i]) {
	    case 1:
	    case 7:
		btn_type = fujitsu_bios[i + 2];
		break;

	    case 4:
		led_type = fujitsu_bios[i + 2];
		break;
	    default:
#if DEBUG > 0
	printk(KERN_NOTICE"fjkeyinf: unsupported device %d.\n",fujitsu_bios[i]);
#endif
		break;
	}
    }

#if DEBUG > 1
    printk(KERN_DEBUG "fjkeyinf: Configuration block [%c%c%c%c%c%c%c%c] "
			"(%d, %d, %d, 0x%02x) (%d, %d, %d, 0x%02x)\n",
			fujitsu_bios[0], fujitsu_bios[1],
			fujitsu_bios[2], fujitsu_bios[3],
			fujitsu_bios[4], fujitsu_bios[5],
			fujitsu_bios[6], fujitsu_bios[7],
			(int)fujitsu_bios[8], (int)fujitsu_bios[9],
			(int)fujitsu_bios[10], (int)fujitsu_bios[11],
			(int)fujitsu_bios[12], (int)fujitsu_bios[13],
			(int)fujitsu_bios[14], (int)fujitsu_bios[15]);

    printk(KERN_DEBUG "fjkeyinf: device found...\n");
#endif
	
    i2c_add_driver(&fjkeyinf_driver);
    return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
    i2c_del_driver(&fjkeyinf_driver);
}
#endif
